import os
import random
import shutil

# Directory containing the images to divide
source_directory = r"D:\AMY\01_09_2K3_TWF_ORIG_TEST\TWF_Process - Copy"
# List of 5 directories to create for the random assignments
output_directories = [r"D:\AMY\01_09_2K3_TWF_ORIG_TEST\230912TWF_process_run1", r"D:\AMY\01_09_2K3_TWF_ORIG_TEST\230912TWF_process_run2", r"D:\AMY\01_09_2K3_TWF_ORIG_TEST\230912TWF_process_run3", r"D:\AMY\01_09_2K3_TWF_ORIG_TEST\230912TWF_process_run4", r"D:\AMY\01_09_2K3_TWF_ORIG_TEST\230912TWF_process_run5"]

# Create the output directories if they don't exist
for directory in output_directories:
    if not os.path.exists(directory):
        os.makedirs(directory)

# List all files in the source directory
all_files = os.listdir(source_directory)

# Shuffle the list of files randomly
random.shuffle(all_files)

# Calculate the number of files to place in each directory
num_files_per_directory = len(all_files) // len(output_directories)

# Distribute the files evenly among the output directories
for i, output_directory in enumerate(output_directories):
    start_index = i * num_files_per_directory
    end_index = (i + 1) * num_files_per_directory if i < len(output_directories) - 1 else len(all_files)
    files_to_copy = all_files[start_index:end_index]

    for filename in files_to_copy:
        source_path = os.path.join(source_directory, filename)
        output_path = os.path.join(output_directory, filename)
        shutil.copy(source_path, output_path)
        print(f"Copied {filename} to {output_directory}")

print("Copying completed.")



### Import libraries ###


import pickle
import os
import multiprocessing as mp


import tqdm
import scipy.io as sio
import numpy as np
import tifffile
import zarr
from PIL import Image

### Import libraries ###



def printProgressBar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill=chr(0x00A3)):#'='): #chr(0x00A3)
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration / total)
    bar = fill * filledLength + '>' + '.' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end="")
    # Print New Line on Complete
    if iteration == total:
        print()


class GEN_TILES(object):
    def __init__(self,
                 input_dir,
                 output_dir,
                 ext,
                 num_processes,
                 exp_dir,
                 objective_power,
                 slide_dimension,
                 rescale,
                 filename,
                 tiles_objective_value,
                 tiles_read_size
                 ):
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.ext = ext
        self.num_processes = num_processes
        self.objective_power = objective_power
        self.tiles_read_size = tiles_read_size
        self.rescale = rescale

        if os.path.exists(self.output_dir) is False:
            os.makedirs(self.output_dir)

    def generate_tiles(self, slide_names_name_list, process_num):

        for s_n, slide_name in enumerate(slide_names_name_list):

            print('Process number:{}...Creating patches from slide {}... {}/{}'.format(process_num, slide_name, s_n,
                                                                                       process_num))

            if os.path.exists(
                    os.path.join(self.output_dir, os.path.basename(slide_name))) is False:
                os.mkdir(os.path.join(self.output_dir, os.path.basename(slide_name)))

            img = tifffile.imread(os.path.join(self.input_dir, slide_name), aszarr=True)
            img_data = zarr.open(img, 'r')
            #
            h, w, c = img_data[0].shape
            print(h, w, c)
            k = 0
            #
            PATCH_SIZE = self.tiles_read_size[0]

            img_num = int(np.ceil(h / PATCH_SIZE) * np.ceil(w / PATCH_SIZE))
            ### Image tiling code ###
            for i in range(0, h, 2000):
                for j in range(0, w, 2000):

                    if (j + 2000 > w) and (i + 2000 <h):
                        out = img_data[0][i:i +PATCH_SIZE, j:w]
                    if (i + 2000 > h) and (j + 2000 < w):
                        out = img_data[0][i:h, j:j + PATCH_SIZE]
                    if (i + 2000 > h) and (j + 2000 > w):
                        out = img_data[0][i:h, j:w]
                    if (i + 2000 < h) and (j + 2000 < w):
                        out = img_data[0][i:i + PATCH_SIZE, j:j + PATCH_SIZE]

                    Image.fromarray(out).save(os.path.join(self.output_dir,
                                                           os.path.basename(slide_name),
                                                           f"{'Da'}{k}{'.jpg'}"))

                    k += 1

            printProgressBar(img_num, len(range(0, img_num)), prefix='Progress:',
                             suffix='Completed and created total number of patches = %s' % img_num, length=5)



    # Multi processing by taking into consideration
    # the number of processes.

    def generate_params(self):

        for slide_name in os.listdir(self.input_dir):
            if not slide_name.endswith('.ndpi'):
                continue
            if os.path.exists(
                    os.path.join(self.output_dir, os.path.basename(slide_name))) is False:
                os.mkdir(os.path.join(self.output_dir, os.path.basename(slide_name)))
            params = {}
            img = tifffile.imread(os.path.join(self.input_dir, slide_name), aszarr=True)
            img_data = zarr.open(img, 'r')

            h, w, c = img_data[0].shape

            # Generate thumbnail image
            thumbx, thumby, thumbc = h // 16, w // 16, c

            # print(thumbx, thumby)
            thumb = img_data[4][0:thumbx, 0:thumby, :]

            print(thumb.shape)
            Image.fromarray(thumb).save(os.path.join(self.output_dir, os.path.basename(slide_name),
                                                     os.path.basename(slide_name) +
                                                     '_thumbnail.jpg'))

            with tifffile.TiffFile(os.path.join(self.input_dir, slide_name)) as f:
                XRES = f.pages[0].tags['XResolution'].value
                YRES = f.pages[0].tags['YResolution'].value
                RESUNIT = f.pages[0].tags['ResolutionUnit'].value

            h, w, c = img_data[0].shape
            dimension_y, dimension_x, dim_c = w, h, c
            params['slide_dimension'] = (dimension_y, dimension_x)
            params['exp_dir'] = os.path.join(self.output_dir, slide_name)
            params['filename'] = slide_name
            params['XRES'] = XRES
            params['YRES'] = YRES
            params['RESUNIT'] = RESUNIT
            params['rescale'] = self.rescale
            params['tiles_read_size'] = self.tiles_read_size
            with open(os.path.join(self.output_dir, os.path.basename(slide_name), 'param.p'),
                      'wb') as file:
                pickle.dump(params, file)
                sio.savemat(os.path.join(self.output_dir, os.path.basename(slide_name),
                                         'param.mat'), params)

    def apply_multiprocessing(self):
        l = [fname for fname in os.listdir(self.input_dir) if fname.endswith(self.ext) is True]
        n = len(l)
        num_elem_per_process = int(np.ceil(n / int(self.num_processes)))

        file_names_list_list = []

        for i in range(int(self.num_processes)):
            start_ = i * num_elem_per_process
            x = l[start_: start_ + num_elem_per_process]
            file_names_list_list.append(x)

        print('{} processes created.'.format(int(self.num_processes)))
        # create list of processes
        processes = [
            mp.Process(target=self.generate_tiles, args=(file_names_list_list[process_num], process_num)) for
            process_num in range(int(self.num_processes))]

        # Run processes
        for p in processes:
            p.start()

            printProgressBar(0, int(self.num_processes), prefix='Progress:', suffix="Complete\n", length=1)

        # Exit the completed processes
        for p in processes:
            p.join()
        print('All Processes finished!!!')

    def run(self):

        if int(self.num_processes) == 1:
            file_names_list = [fname for fname in os.listdir(self.input_dir) if fname.endswith(self.ext) is True]
            self.generate_tiles(file_names_list, 1)
        else:
            self.apply_multiprocessing()


def parse_options(opts_in):
    # params = opts_in
    params_keys = {'slides': 'input_dir', 'patches': 'output_dir', 'ext': 'ext', 'nump': 'num_processes'}

    params = {params_keys[key]: value for key, value in opts_in.items()}

    params.update({'exp_dir': '', 'objective_power': 20, 'slide_dimension': [],
                   'rescale': 1, 'filename': [], 'tiles_objective_value': 20,
                   'tiles_read_size': (2000, 2000)})

    print(params)

    obj = GEN_TILES(**params)
    obj.run()
    obj.generate_params()


def perform_tiling(slide_directory, output_directory, extension, num_processes):
      """Perform tile extraction given a directory of slides, an output directory,
      extension of the slides, and the number of the processes to be used for tiling.

      Parameters
      ----------
      slide_directory : Path
        Path to the slides directory
      output_directory : Path
        Path to the patches saved in corresponding slide directory
      extension : {'.ndpi', '.svs', '.qptiff', '.mrxs', '.tif', '.tiff', '.scn'}
        Type of the slide or custom scanner extension
      num_processes : int
        Number of process needed to run on the CPU

      Returns
      -------
      None
      """
      params_keys = {'input_dir': slide_directory,
                     'output_dir': output_directory,
                     'ext':extension,
                     'num_processes': num_processes}

      params_keys.update({'exp_dir': '', 'objective_power': 20, 'slide_dimension': [],
                       'rescale':1, 'filename': [], 'tiles_objective_value': 20,
                        'tiles_read_size': (2000, 2000)})


      obj = GEN_TILES(**params_keys)
      obj.run()
      obj.generate_params()


# slides = r'D:\Projects\HE_Int_ts_cc\Sarcoma\Well_Diff\slides\new_slides'
# output = r'D:\Projects\HE_Int_ts_cc\Sarcoma\Well_Diff\tiles'
#
# #perform_tiling('./slides', './output', '.ndpi', 1)
# perform_tiling(slides, output, '.ndpi', 1)




import os
import shutil

cws_path=r'D:\AMY\230803_PSR_ROI_tiles\cws'


copy_ss1_path =r'E:\CODE\CD-31 staining quantization project\stain_mask'
name=[]

for slide in os.listdir(cws_path):
    if slide.endswith('.ndpi'):
        try:
            for roi_dir in os.listdir(os.path.join(cws_path, slide, 'ROI_TWF')):
                if os.path.isdir(os.path.join(cws_path, slide, 'ROI_TWF', roi_dir)):
                    if os.path.exists(os.path.join(copy_ss1_path, slide, roi_dir)) is False:
                        os.makedirs(os.path.join(copy_ss1_path, slide, roi_dir))
                    #print(roi_dir)
                    for Da_img in os.listdir(os.path.join(cws_path, slide, 'ROI_TWF', roi_dir)):

                        if Da_img.endswith('_smooth.jpg'):
                            # print(img)

                            # imgname=img.split('.ndpi_Ss1.jpg')[0]
                            # name.append(imgname)
                            src = os.path.join(cws_path, slide, 'ROI_TWF', roi_dir, Da_img)
                            dst = os.path.join(copy_ss1_path, slide, roi_dir, Da_img)
                            #print(dst)
                            shutil.copy(src, dst)
                            #srcname = Da_img
                            print(dst)
                    # print(srcname)
                    # copyfile(os.path.join(cws_path, slide,srcname),dst)
        except Exception:
            print(slide)
            pass



import numpy as np
import cv2
import math
from scipy.ndimage import binary_fill_holes
def calculate_magnitude(od):

    channel = cv2.split(od)
    square_b = cv2.pow(channel[0], 2)
    square_g = cv2.pow(channel[1], 2)
    square_r = cv2.pow(channel[2], 2)
    square_bgr = square_b + square_g + square_r
    magnitude = cv2.sqrt(square_bgr)

    return magnitude

def normaliseOD(od,magnitude):

    channels=cv2.split(od)#, (channels[0],channels[1],channels[2]))

    od_norm_b = cv2.divide(channels[0], magnitude)
    od_norm_g = cv2.divide(channels[1], magnitude)
    od_norm_r = cv2.divide(channels[2], magnitude)

    od_norm = cv2.merge((od_norm_b,od_norm_g, od_norm_r))

    return od_norm

#remove artefacts based off optical density e.g edge of slide
def Artifact_SS1(img):
    I = img.transpose()
    k, width, height = I.shape
    I = I.reshape(k, width * height)
    I = np.float32(I)

    od = cv2.max(I, 1)

    grey_angle = 0.2

    magnitude_threshold = 0.05

    channels = cv2.split(od)
    #
    magnitude = np.zeros(od.shape)
    #
    background = 245
    #
    # Convert channels to a list to allow modification
    channels = list(channels)

    # Perform the operation
    background = 245
    channels[0] /= background

    # Convert the list back to a tuple
    channels = tuple(channels)
    # channels[0]/=background # old code line converted to suppport the processing
    # Extract the array from the tuple
    # image_array = channels[0]
    # # Modify the first channel (index 0)
    # image_array[0] /= background
    # # If you need to keep it as a tuple, you can recreate it
    # channels = (image_array,)
    # Assuming 'channels' is a tuple of numpy arrays
    # channels = list(channels)  # Convert tuple to list
    # channels[0] = channels[0] / background  # Use numpy array division
    # channels = tuple(channels)

    od = cv2.merge(channels)

    od = cv2.log(od)

    od *= (1 / cv2.log(10)[0])

    od = -od
    od = od.reshape(3, width, height).transpose()
    magnitude = calculate_magnitude(od)

    tissue_and_artefact_mask = (magnitude > magnitude_threshold)

    od_norm = normaliseOD(od, magnitude)

    chan = cv2.split(od_norm)

    grey_mask = (chan[0] + chan[1] + chan[2]) >= (math.cos(grey_angle) * cv2.sqrt(3)[0])

    other_colour_mask = (chan[2] > chan[1]) | (chan[0] > chan[1])

    mask = grey_mask | other_colour_mask

    mask = (255 - mask) & tissue_and_artefact_mask
    mask1 = mask.astype(np.int8)

    mask_img = mask1*255
    mask_img = mask_img.astype(np.uint8)

    mask_img_smooth = cv2.GaussianBlur(mask_img, (151, 151), 0)
    mask_img_smooth = cv2.threshold(mask_img_smooth, 20, 255, cv2.THRESH_BINARY)[1]

    mask_img_smooth = binary_fill_holes(mask_img_smooth)
    mask_img_smooth = mask_img_smooth.astype(np.uint8)*255


    clean = cv2.bitwise_and(img, img, mask=mask1)
    clean = cv2.bitwise_not(clean)

    clean = cv2.bitwise_and(clean, clean, mask=mask1)
    clean = cv2.bitwise_not(clean)

    img_s = cv2.cvtColor(clean, cv2.COLOR_BGR2GRAY)
    img_th = cv2.threshold(img_s, 120, 255, cv2.THRESH_BINARY)
    #print(img_th)

    #write_mask1 = mask.astype(np.uint8)*255

    #cv2.imwrite(os.path.join(output_dir, os.path.splitext(im)[0]+'_Mask.jpg'), write_mask1)
    return clean, mask_img_smooth



import napari
from magicgui import magicgui, widgets
from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QStackedWidget, QLabel, QTabWidget
from skimage import color, filters
import numpy as np
import sys
import pathlib

### QWidget implementation from napIPUvis and napari-tma github references ###
#P-MAT - a napari-based workflow for scale-independent quantification of the extracellular matrix and its topological scale-independent descriptors
'''
https://github.com/IntegratedPathologyUnit-ICR/POPIDD/tree/napari_tma

'''
# Usage in napari

def main():
    viewer = napari.Viewer()

    @magicgui(
            input_slide_dir={"widget_type": "FileEdit", "mode": "d"},
            wsi_tiles_dir={"widget_type": "FileEdit", "mode": "d"},
            file_type={"widget_type": "FileEdit", "mode": "r"},
            num_process={"widget_type": "ComboBox", "choices": ["1", "2", "4", "8"]},
            call_button="Run"  # This adds a Run button to the widget
            )
    def wGTauto(
            input_slide_dir=pathlib.Path("slides"),
            wsi_tiles_dir=pathlib.Path("cws"),
            file_type=pathlib.Path(".ndpi"),
            num_process="1",
            # loadMaskTS=bool
            ):
        try:
            from utils_PMAT import generate_tiles_from_PSR_wsi_slides
            generate_tiles_from_PSR_wsi_slides.perform_tiling(str(input_slide_dir), str(wsi_tiles_dir), str(file_type), num_process)
            print("Extraction completed!")
            print(f"Parameters: {input_slide_dir}, {wsi_tiles_dir}, {file_type}, {num_process}")
        except ModuleNotFoundError as e:
            print(f"Error: Could not import from PMAT module. {e}")
            print(f"Python path: {sys.path}")
        except Exception as e:
            print(f"An error occurred: {e}")

    @magicgui(
            input_slide_dir={"widget_type": "FileEdit", "mode": "d"},
            wsi_tiles_dir={"widget_type": "FileEdit", "mode": "d"},
            output_dir={"widget_type": "FileEdit", "mode": "d"},
            file_type={"widget_type": "ComboBox", "choices": [".svs", ".ndpi", ".tif"]},
            thresh_d={"widget_type": "ComboBox", "choices": ["80", "60", "40", "20"]},
            call_button="Run"  # This adds a Run button to the widget
            )
    def wTSauto(
            input_slide_dir=pathlib.Path("slides"),
            wsi_tiles_dir=pathlib.Path("cws"),
            output_dir=pathlib.Path("refined_workflow"),
            file_type=".ndpi",
            thresh_d="80",
            ):
        try:
            from utils_PMAT import extract_ROIs_from_annotations
            extract_ROIs_from_annotations.map_of_slides_and_annotations(str(input_slide_dir), str(wsi_tiles_dir), str(output_dir), file_type, int(thresh_d))
            print("Extraction completed!")
            print(f"Parameters: {input_slide_dir}, {wsi_tiles_dir}, {output_dir}, {file_type}, {(thresh_d)}")
        except ModuleNotFoundError as e:
            print(f"Error: Could not import from PMAT module. {e}")
            print(f"Python path: {sys.path}")
        except Exception as e:
            print(f"An error occurred: {e}")

    @magicgui(
        cws_dir={"widget_type": "FileEdit", "mode": "d"},
        annot_dir={"widget_type": "FileEdit", "mode": "d"},
        output_low_res_dir={"widget_type": "FileEdit", "mode": "d"},
        high_res_dir={"widget_type": "FileEdit", "mode": "d"},
        specific_dir={"widget_type": "ComboBox",
                      "choices": ["ROI_80_AFC", "ROI_80_H1", "ROI_80_PS1", "ROI_60_AFC", "ROI_60_H1", "ROI_60_PS1",
                                  "ROI_40_AFC", "ROI_40_H1", "ROI_40_PS1"]},
        scale={"widget_type": "ComboBox", "choices": ["16", "8", "4"]},
        call_button="Run"  # This adds a Run button to the widget
    )
    def wSTauto(
            cws_dir=pathlib.Path("cws_folder"),
            annot_dir=pathlib.Path("refined_workflow"),
            output_low_res_dir=pathlib.Path("output_LRES_dir"),
            high_res_dir=pathlib.Path("output_HRES_dir"),
            specific_dir="ROI_40_H1",
            scale="16",

    ):
        try:
            from utils_PMAT import stitch_Low_res_per_ROI_with_80percent_tile_selection_TWF
            stitch_Low_res_per_ROI_with_80percent_tile_selection_TWF.get_SS1_dimension_image_from_cws_resolution(
                str(cws_dir), str(annot_dir), str(output_low_res_dir), str(high_res_dir), str(specific_dir), int(scale))

        except ModuleNotFoundError as e:
            print(f"Error: Could not import from PMAT module. {e}")
            print(f"Python path: {sys.path}")
        except Exception as e:
            print(f"An error occurred: {e}")

    @magicgui(
        wsi_tiles_dir={"widget_type": "FileEdit", "mode": "d"},
        output_dir={"widget_type": "FileEdit", "mode": "d"},
        file_type={"widget_type": "ComboBox", "choices": [".svs", ".ndpi", ".tif"]},
        shape_type={"widget_type": "ComboBox", "choices": ["Tree", "Branch"]},
        call_button="Run"  # This adds a Run button to the widget
    )
    def wSFauto(
            wsi_tiles_dir=pathlib.Path("cws"),
            output_dir=pathlib.Path("refined_workflow"),
            file_type=".ndpi",
            shape_type="Tree",
    ):
        try:
            from utils_PMAT import extract_ROIs_from_annotations
            extract_ROIs_from_annotations.map_of_slides_and_annotations(str(wsi_tiles_dir), str(output_dir), file_type, str(shape_type))
            print("Extraction completed!")
            print(f"Parameters: {wsi_tiles_dir}, {output_dir}, {file_type}, {shape_type}")
        except ModuleNotFoundError as e:
            print(f"Error: Could not import from PMAT module. {e}")
            print(f"Python path: {sys.path}")
        except Exception as e:
            print(f"An error occurred: {e}")

    tab_widget = QTabWidget()
    #

    tab_GT = QWidget()#tile
    gt_layout = QVBoxLayout()
    tab_GT.setLayout(gt_layout)
    gt_layout.addWidget(wGTauto.native)


    # %%
    tab_TS = QWidget() #Annotated tissue ROI selection
    ts_layout = QVBoxLayout()
    tab_TS.setLayout(ts_layout)
    ts_layout.addWidget(wTSauto.native)

    tab_ST = QWidget()#stitch
    st_layout = QVBoxLayout()
    tab_ST.setLayout(st_layout)
    st_layout.addWidget(wSTauto.native)

    tab_SF = QWidget()  # stitch
    sf_layout = QVBoxLayout()
    tab_SF.setLayout(sf_layout)
    sf_layout.addWidget(wSFauto.native)
    # %%
    tab_widget.addTab(tab_GT, "Generate tiles from WSIs")
    tab_widget.addTab(tab_TS, "Generate ROI regions of annotation")
    tab_widget.addTab(tab_ST, "Stitch processed outputs")
    tab_widget.addTab(tab_SF, "Feature descriptor extraction")

    # Connect the update button to the update_info method
    #tab_widget.update_info_btn.clicked.connect(tab_widget.update_info)

    tab_widget.setFixedHeight(500)
    # tab_widget1.setFixedHeight(400)
    viewer.window.add_dock_widget(
        tab_widget, name="ECM Process pipeline", area="right",
        add_vertical_stretch=False, tabify=True)


    #

    napari.run()
if __name__=='__main__':
    main()








import os
import numpy as np
import glob
import re
import cv2

# cws_folder = os.path.normpath(r'D:\AMY\230807_thumbnail_img_metadata')
# output_dir = r'D:\AMY\230803_PSR_ROI_tiles\cws'


def natural_key(string_):

    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)]

def dark_pink_filter(cws_folder, output_dir):

    wsi_files = sorted(glob.glob(os.path.join(cws_folder, '*.ndpi')))

    # if os.path.exists(output_dir) is False:
    #     os.makedirs(output_dir)

    for wsi in range(0, len(wsi_files)):

        filename = wsi_files[wsi]
        drivepath, slidename = os.path.split(wsi_files[wsi])

        # if os.path.exists(os.path.join(output_dir, filename)) is False:
        #     os.makedirs(os.path.join(output_dir, filename))

         #if not os.path.exists(os.path.join(self.output_dir, cws_dir, slide)):

        if os.path.exists(os.path.join(output_dir, filename, "ROI_TWF_FILTER_ORIG")) is False:
            os.makedirs(os.path.join(output_dir, filename, "ROI_TWF_FILTER_ORIG"))

        if os.path.exists(os.path.join(output_dir, filename, "ROI_TWF_FILTER_INTER")) is False:
            os.makedirs(os.path.join(output_dir, filename, "ROI_TWF_FILTER_INTER"))



        print(os.path.join(output_dir, slidename, "ROI_with_AFC"))

        annotated_dir_i = os.path.join(output_dir, slidename, 'ROI_with_AFC') ######### New directory mentioned to actively take only the .jpg files here ######
        images = sorted(os.listdir(annotated_dir_i), key=natural_key)

        for ii in images:
            #ri=[]
            imagelist =[]
            print(ii)
            if os.path.isdir(os.path.join(output_dir, slidename, 'ROI_with_AFC', ii)):

                if os.path.exists(os.path.join(output_dir, slidename, "ROI_TWF_FILTER_ORIG", ii)) is False:
                    os.makedirs(os.path.join(output_dir, slidename, "ROI_TWF_FILTER_ORIG", ii))

                if os.path.exists(os.path.join(output_dir, filename, "ROI_TWF_FILTER_INTER", ii)) is False:
                    os.makedirs(os.path.join(output_dir, filename, "ROI_TWF_FILTER_INTER", ii))
                 #

                #print(len(os.listdir(os.path.join(annotated_dir, imagename, 'ROI_80_H', ii))))
                for roi_i in os.listdir(os.path.join(output_dir, slidename, 'ROI_with_AFC', ii)):
                    #print(roi_i)
                    roi_name = roi_i
                    #print(roi_name)

                    if roi_name.endswith('.jpg'):
                        if os.path.isfile(os.path.join(output_dir, slidename, 'ROI_with_AFC', ii, roi_name)):

                            orig_PS = cv2.imread(os.path.join(output_dir, slidename, 'ROI_80_PS1', ii, roi_name))
                            bin_mask = cv2.cvtColor(orig_PS, cv2.COLOR_BGR2GRAY)
                            #bin_mask = cv2.threshold(bin_mask, 40, 255, cv2.THRESH_BINARY_INV)[1]


                            # Set the threshold value
                            threshold_value = 100 ####

                            # Create a mask of pixels less than the threshold value
                            mask = bin_mask < threshold_value
                            mask = mask*255
                            mask = np.uint8(mask)
                            ### Add/Modify to 3channel if the mask is two channel
                            mask = cv2.cvtColor(mask, cv2.COLOR_GRAY2RGB)

                            da_and_img = cv2.bitwise_and(orig_PS, mask)
                            invert_mask = cv2.bitwise_not(mask)
                            roi_with_bg_img = cv2.bitwise_or(da_and_img, invert_mask)


                            # Define kernel for morphological operations
                            #kernel = np.ones((15, 15), np.uint8)

                            # Perform closing operation to remove small black noise pixels
                            #closed_image = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel)

                            # Perform opening operation to remove small white noise pixels
                            #opened_image = cv2.morphologyEx(closed_image, cv2.MORPH_OPEN, kernel)
                            # Apply the mask to set the selected pixels to the desired value (255)
                            #thresholded_image = np.where(mask, 255, bin_mask)
                            #
                            #
                            #bin_image_smooth = cv2.GaussianBlur(opened_image, (21, 21), 0)

                            cv2.imwrite(os.path.join(output_dir, slidename, "ROI_TWF_FILTER_INTER", ii, os.path.splitext(roi_name)[0]+".png"), da_and_img)
                            cv2.imwrite(os.path.join(output_dir, slidename, "ROI_TWF_FILTER_ORIG", ii, os.path.splitext(roi_name)[0]+".png"), roi_with_bg_img)

                            print(os.path.join(output_dir, slidename, "ROI_TWF_FILTER_ORIG", ii, os.path.splitext(roi_name)[0]+".png"))



        #filename = wsi_files[wsi]

if __name__ == '__main__':

    params = {
                'cws_folder': os.path.normpath(r'D:\Projects\cbias-nap-AMY\cws1'),
                'output_dir': r'D:\Projects\cbias-nap-AMY\refined_workflow\cws', #r'D:\AMY\230803_PSR_ROI_tiles\cws',
                'refine_mask_dir': r'D:\AMY\230810_low_res_mask_per_ROI',
                'scale': 16
               }
    dark_pink_filter(params['cws_folder'], params['output_dir'])



from setuptools import setup, find_packages

setup(
    name="c-PMAT",
    version="0.2.0",
    author="Priya Narayanan",
    author_email="priya.narayanan@icr.ac.uk",
    description="Quantification of extra cellular matrix from whole slide images through preprocessing using PMAT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/yourusername/your-repo",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=[
        "napari==0.4.19",
        "napari-console==0.0.9",
        "napari-ome-zarr==0.5.2",
        "napari-plugin-engine==0.2.0",
        "napari-plugin-manager==0.1.0a2",
        "napari-svg==0.1.10",
        "napari-wsi==0.1.4",
        "pillow==10.2.0",
        "imagecodecs==2024.1.1",
        "imageio==2.34.0",
        "imagesize==1.4.1",
        "numba==0.59.0",
        "numcodecs==0.12.1",
        "numpy==1.26.4",
        "numpydoc==1.6.0",
        "oauthlib==3.2.2",
        "ome-zarr==0.8.3",
        "opencv-python==3.4.16.59",
        "openslide-python==1.3.1",
        "PyQt5==5.15.10",
        "PyQt5-Qt5==5.15.2",
        "PyQt5-sip==12.13.0",
        "QtPy==2.4.1",
        "rasterio==1.3.9",
        "scikit-image==0.22.0",
        "scipy==1.12.0",
        "setuptools==68.0.0",
        "tifffile==2024.2.12",
        "zarr==2.17.0"


    ],
)


##################### batch for images per ROI ################



import pickle
import os
import glob
import cv2
import re

import numpy as np



def natural_key(string_):

    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)]

def printProgressBar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='='): #chr(0x00A3)
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration / total)
    bar = fill * filledLength + '>' + '.' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end="")
    # Print New Line on Complete
    if iteration == total:
        print()

def get_SS1_dimension_image_from_cws_resolution(cws_folder,annotated_dir,output_dir, refine_mask_dir, specific_dir, scale):

    wsi_files = sorted(glob.glob(os.path.join(cws_folder, '*.ndpi')))

    if os.path.exists(output_dir) is False:
        os.makedirs(output_dir)

    if os.path.exists(refine_mask_dir) is False:
        os.makedirs(refine_mask_dir)

    for wsi in range(0, len(wsi_files)):

        filename = wsi_files[wsi]


        param = pickle.load(open(os.path.join(filename, 'param.p'), 'rb'))

        slide_dimension = np.array(param['slide_dimension']) / 1 #1param['rescale']

        slide_w, slide_h = slide_dimension
        print(slide_w, slide_h)
        cws_w, cws_h = 2000, 2000 #param['cws_read_size']

        divisor_w = np.ceil(slide_w / cws_w)
        divisor_h = np.ceil(slide_h / cws_h)

        w, h = int(slide_w / scale), int(slide_h / scale)
        print('%s, Ss1 size: %i,%i'%(os.path.basename(filename), w, h))


        drivepath, imagename = os.path.split(wsi_files[wsi])

        if os.path.exists(os.path.join(output_dir, imagename)) is False:
            os.makedirs(os.path.join(output_dir, imagename))

        if os.path.exists(os.path.join(refine_mask_dir, imagename)) is False:
            os.makedirs(os.path.join(refine_mask_dir, imagename))

        ######### New directory mentioned to actively take only the .jpg files here ######

        annotated_dir_i = os.path.join(annotated_dir, imagename, specific_dir)
        images = sorted(os.listdir(annotated_dir_i), key=natural_key)

        img_all = np.zeros((h, w, 3))
        img_all = img_all.astype(np.uint8)

        img_high_res_all = np.zeros((int(slide_h), int(slide_w), 3))
        img_high_res_all = img_high_res_all.astype(np.uint8)
        scale_0 = 1

        for ii in images:
            #ri=[]
            imagelist =[]
            if os.path.isdir(os.path.join(annotated_dir, imagename, specific_dir, ii)):



                ## ii is checked for being a directory or not and then its created in the output directory for each slide
                if os.path.exists(os.path.join(output_dir, imagename, ii)) is False:
                    os.makedirs(os.path.join(output_dir, imagename, ii))

                if os.path.exists(os.path.join(refine_mask_dir, imagename, ii)) is False:
                    os.makedirs(os.path.join(refine_mask_dir, imagename, ii))
                 #

                #print(len(os.listdir(os.path.join(annotated_dir, imagename, 'ROI_80_H', ii))))
                for roi_i in glob.glob(os.path.join(annotated_dir, imagename, specific_dir, ii, '*.jpg')):
                    #print(roi_i)
                    roi_name = os.path.basename(roi_i)

                    if roi_name.endswith('.jpg'):
                        new_path = roi_name
                        print(os.path.join(annotated_dir, imagename, specific_dir, ii, roi_name))
                        if os.path.isfile(os.path.join(annotated_dir, imagename, specific_dir, ii, roi_name)):
                            imagelist.append(new_path)
                print(imagelist)#   each ROI da image list belonging to individual slide

                imagelist_roi = sorted(imagelist, key=natural_key)
                if len(imagelist)==0: ################################ handle empty roi directory
                    continue
                printProgressBar(0, len(imagelist), prefix='Progress:', suffix='Complete', length=50)

                for i in imagelist_roi:
                    cws_i = int(re.search(r'\d+', i).group())
                    h_i = np.floor(cws_i / divisor_w) * cws_h
                    w_i = (cws_i - h_i / cws_h * divisor_w) * cws_w

                    h_0_i = int(h_i / scale_0)
                    w_0_i = int(w_i / scale_0)

                    h_i = int(h_i / scale)
                    w_i = int(w_i / scale)
                    print(h_i, w_i)



                    # print(cws_i, w_i, h_i)

                    img_H = cv2.imread(os.path.join(annotated_dir, imagename, specific_dir, ii, i))
                    # img_H = cv2.cvtColor(img_H, cv2.COLOR_BGR2GRAY)
                    #img = cv2.imread(os.path.join(annotated_dir, imagename, i))
                    #im_mask = cv2.imread(os.path.join(annotated_dir, imagename, 'img_mask', ii, i))
                    #da_and_img = cv2.bitwise_and(img, im_mask)
                    #mask_artifact_corrected = Artifact_SS1(img)[1]
                    #total_pixels = mask_artifact_corrected.shape[0] * mask_artifact_corrected.shape[1]

                    # Count the number of white pixels
                    #white_pixels = np.count_nonzero(mask_artifact_corrected)

                    #Calculate the percentage of white pixels
                    #white_percentage = (white_pixels / total_pixels) * 100
                    #print(int(white_percentage))

                    # if white_percentage < 80:
                    #     continue

                    img = cv2.resize(img_H, (int(img_H.shape[1]/scale), int(img_H.shape[0]/scale)))



                    img_all[h_i : h_i + int(img.shape[0]), w_i : w_i + int(img.shape[1])] = img

                    img_high_res_all[h_0_i : h_0_i + int(img_H.shape[0]), w_0_i : w_0_i + int(img_H.shape[1])] = img_H

                    #img_H = cv2.resize(img_H, (int(img_H.shape[1]/scale), int(img_H.shape[0]/scale)))

                   # img_all_H[h_i : h_i + int(img.shape[0]), w_i : w_i + int(img.shape[1]), :] = img_H

                    #img_refine = cv2.resize(da_and_img, (int(da_and_img.shape[1]/scale), int(da_and_img.shape[0]/scale)))

                    #img_all_refine[h_i : h_i + int(img.shape[0]), w_i : w_i + int(img.shape[1]), :] = img_refine


                    printProgressBar(cws_i, len(images), prefix='Progress:',
                                     suffix='Completed for %s'%i, length=50)

                    #if w_i + cws_w / scale > w:
                ############# refine Da's for visualisation ######################

                #bm_img_roi_low_res = cv2.imread(os.path.join(refine_mask_dir, imagename, ii, imagename+'_'+ii+".png"))
                # BCPP PSR B22-966 18565-10-A2 s7 batch 2.ndpi ### ROI 4 is empty #####
               # if bm_img_roi_low_res is None: #################### if the ROI is None, then this if condition helps to exclude
                   # continue

              #  result_img = cv2.bitwise_and(np.uint8(img_all_H), bm_img_roi_low_res)
              # invert_mask = cv2.bitwise_not(bm_img_roi_low_res)
               # roi_with_bg_img = cv2.bitwise_or(result_img, invert_mask)

                #
                ############## refine Da's for visualisation ######################
                cv2.imwrite(os.path.join(output_dir, imagename, ii, os.path.splitext(imagename)[0]+'_'+ii+"_AFC.png"), img_all)
                cv2.imwrite(os.path.join(refine_mask_dir, imagename, ii, os.path.splitext(imagename)[0]+'_'+ii+"_PS_AFC.png"),img_high_res_all)

                #cv2.imwrite(os.path.join(output_dir, imagename, ii, os.path.splitext(imagename)[0]+'_'+ii+"_refined_img.png"), img_all_refine)
                #cv2.imwrite(os.path.join(output_dir, imagename, os.path.splitext(imagename)[0]+'_'+ii+"_Heam.png"), result_img)
                #cv2.imwrite(os.path.join(output_dir, imagename, os.path.splitext(imagename)[0]+'_'+ii+"_Heam_white_bg.png"), roi_with_bg_img)






# if __name__ == '__main__':
#
#     params = {
#                 'cws_folder': os.path.normpath(r'D:\Projects\c-PMAT\src\c-PMAT\cws'),
#                 'annotated_dir': r'D:\Projects\c-PMAT\src\c-PMAT\refined_workflow\cws',
#                 'output_dir': r'D:\Projects\c-PMAT\src\c-PMAT\output_LRES_dir',
#                 'refine_mask_dir': r'D:\Projects\c-PMAT\src\c-PMAT\output_LRES_dir',
#                 'specific_dir':'ROI_40_H1',
#                 'scale': 16
#                }
#
#     get_SS1_dimension_image_from_cws_resolution(params['cws_folder'], params['annotated_dir'],
#                                                 params['output_dir'], params['refine_mask_dir'],
#                                                 params['specific_dir'], params['scale'])



#TITLE: COLOUR NORMALISATION
#AIM: The PSR images encompass a wide range of RGB values, to progress to a one size fits all colour deconvolution we first
# need to normalise the RGB values. To do so, we will extract the RGB values, plot them and take the avergae value to apply to all slides.

import os
import numpy as np
from PIL import Image

def convert_RGB_to_OD(I):
    mask = (I == 0)
    I[mask] = 1

    # return np.maximum(-1 * np.log(I / 255), 1e-6)
    return np.maximum(-1 * np.log(I / 255), np.zeros(I.shape) + 0.1)


def convert_OD_to_RGB(OD):
    #assert OD.min() >= 0, "Negative optical density."

    OD = np.maximum(OD, 1e-6)

    return (255 * np.exp(-1 * OD)).astype(np.uint8)

def normalizeStaining(img, saveFile=None, Io=240, alpha=1, beta=0.15):
    ''' Normalize staining appearence of H&E stained images

    Example use:
        see test.py

    Input:
        I: RGB input image
        Io: (optional) transmitted light intensity

    Output:
        Inorm: normalized image
        H: hematoxylin image
        E: eosin image

    Reference:
        A method for normalizing histology slides for quantitative analysis. M.
        Macenko et al., ISBI 2009
    '''

    HERef = np.array([[0.5626, 0.2159],
                      [0.7201, 0.8012],
                      [0.4062, 0.5581]])



    maxCRef = np.array([1.9705, 1.0308])

    # define height and width of image
    h, w, c = img.shape

    # reshape image
    img = img.reshape((-1, 3))

    # calculate optical density
    OD = -np.log((img.astype(np.float64) + 1) / Io)

    # remove transparent pixels
    ODhat = OD[~np.any(OD < beta, axis=1)]

    # compute eigenvectors
    eigvals, eigvecs = np.linalg.eigh(np.cov(ODhat.T))

    # eigvecs *= -1

    # project on the plane spanned by the eigenvectors corresponding to the two
    # largest eigenvalues
    That = ODhat.dot(eigvecs[:, 1:3])

    phi = np.arctan2(That[:, 1], That[:, 0])

    minPhi = np.percentile(phi, alpha)
    maxPhi = np.percentile(phi, 100 - alpha)

    vMin = eigvecs[:, 1:3].dot(np.array([(np.cos(minPhi), np.sin(minPhi))]).T)
    vMax = eigvecs[:, 1:3].dot(np.array([(np.cos(maxPhi), np.sin(maxPhi))]).T)

    # a heuristic to make the vector corresponding to hematoxylin first and the
    # one corresponding to eosin second
    if vMin[0] > vMax[0]:
        HE = np.array((vMin[:, 0], vMax[:, 0])).T
    else:
        HE = np.array((vMax[:, 0], vMin[:, 0])).T

    # rows correspond to channels (RGB), columns to OD values
    Y = np.reshape(OD, (-1, 3)).T

    # determine concentrations of the individual stains
    C = np.linalg.lstsq(HE, Y, rcond=None)[0]

    # normalize stain concentrations
    maxC = np.array([np.percentile(C[0, :], 99), np.percentile(C[1, :], 99)])
    tmp = np.divide(maxC, maxCRef)
    C2 = np.divide(C, tmp[:, np.newaxis])

    # recreate the image using reference mixing matrix
    Inorm = np.multiply(Io, np.exp(-HERef.dot(C2)))
    Inorm[Inorm > 255] = 254
    Inorm = np.reshape(Inorm.T, (h, w, 3)).astype(np.uint8)

    # unmix hematoxylin and eosin
    H = np.multiply(Io, np.exp(np.expand_dims(-HERef[:, 0], axis=1).dot(np.expand_dims(C2[0, :], axis=0))))
    H[H > 255] = 254
    H = np.reshape(H.T, (h, w, 3)).astype(np.uint8)

    E = np.multiply(Io, np.exp(np.expand_dims(-HERef[:, 1], axis=1).dot(np.expand_dims(C2[1, :], axis=0))))
    E[E > 255] = 254
    E = np.reshape(E.T, (h, w, 3)).astype(np.uint8)



    if saveFile is not None:
        Image.fromarray(Inorm).save(saveFile + '.png')
        Image.fromarray(H).save(saveFile + '_H.png')
        Image.fromarray(E).save(saveFile + '_P.png')
        #Image.fromarray(nuc_RGB).save(saveFile + '_HF.png')

    return Inorm, H, E






import os
import numpy as np
import cv2
OPENSLIDE_PATH = r'C:\Tools\openslide-win64-20231011\bin'

import os
if hasattr(os, 'add_dll_directory'):
    # Windows
    with os.add_dll_directory(OPENSLIDE_PATH):
        import openslide
else:
    import openslide
#
#import openslide
import xml.dom.minidom
import xml.etree.ElementTree as ET
import re

from utils_PMAT.get_bg_correction import *
from utils_PMAT.Colour_norm import *

# from utils_PMAT.get_bg_correction import *
# from utils_PMAT.Colour_norm import *



def natural_key(string_):
    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)]


class GenerateWSIannotation_on_cws(object):

    def __init__(self,
                 input_slide_dir,
                 tiles_dir,
                 output_dir,
                 ext,
                 thresh_d):
        self.input_slide_dir = input_slide_dir
        self.tiles_dir = tiles_dir
        self.output_dir = output_dir
        self.ext = ext
        self.thresh_d = thresh_d
        if os.path.exists(self.output_dir) is False:
            os.makedirs(self.output_dir)

    def generate_patch_of_annotated_tiles(self):

        # Annotated tiles positive images will be saved in img_mask directory
        # Background tiles will be saved in Mat_files directory
        # .mat files from this will be used to generate tf records file

        file_names_list =[fname for fname in os.listdir(self.input_slide_dir) if fname.endswith(self.ext) and (fname.startswith('BCPP') or fname.startswith('RADIO') or fname.startswith('PLUMMB')) is True]

        cws_dir = "cws"

        if os.path.exists(os.path.join(self.output_dir, cws_dir)) is False:
            os.makedirs(os.path.join(self.output_dir, cws_dir))

        for slide in file_names_list:

            print(os.path.join(self.output_dir, cws_dir, slide))
            if not os.path.exists(os.path.join(self.output_dir, cws_dir, slide)):


                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide)) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "img_mask")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "img_mask"))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "ROI_DA")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "ROI_DA"))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "ROI_with_AFC")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "ROI_with_AFC"))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_AFC")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_AFC"))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_H1")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_H1"))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_PS1")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_PS1"))

                if os.path.exists(os.path.join(self.output_dir, cws_dir, slide, "ROI_TWF_FILTER_ORIG")) is False:
                    os.makedirs(os.path.join(self.output_dir, cws_dir, slide, "ROI_TWF_FILTER_ORIG"))

                osr = openslide.OpenSlide(os.path.join(self.input_slide_dir, slide))
                level = 0
                ds = osr.level_downsamples[level]
                w, h = osr.level_dimensions[0]

                mask_path = os.path.join(self.output_dir, cws_dir, slide, "img_mask")
                da_configpath = os.path.join(self.output_dir, cws_dir, slide, "ROI_DA")
                roi_corrected = os.path.join(self.output_dir, cws_dir, slide, "ROI_with_AFC")
                roi_corrected_da_path = os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_AFC")
                H_da_path = os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_H1")
                PS_da_path = os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_PS1")
                norm_da_path = os.path.join(self.output_dir, cws_dir, slide, "ROI_"+str(self.thresh_d)+"_norm1")
                twf_da_path = os.path.join(self.output_dir, cws_dir, slide, "ROI_TWF_FILTER_ORIG")


                doc = xml.dom.minidom.parse(os.path.join(self.input_slide_dir, os.path.splitext(slide)[0] + '.xml'))
                Region = doc.getElementsByTagName("Region")

                class_types_reg = []
                for Reg in Region:
                    class_types_reg.append(Reg.getAttribute("Text"))
                class_types = np.unique(class_types_reg)
                for cls_type in class_types:
                    mask_path_cls_type = os.path.join(mask_path, cls_type)
                    roi_path_cls_type = os.path.join(da_configpath, cls_type)
                    roi_corrected_cls_type = os.path.join(roi_corrected, cls_type)
                    roi_80_afc_cls_type = os.path.join(roi_corrected_da_path, cls_type)
                    roi_80_H_cls_type = os.path.join(H_da_path, cls_type)
                    roi_80_PS_cls_type = os.path.join(PS_da_path, cls_type)
                    roi_80_norm_cls_type = os.path.join(norm_da_path, cls_type)
                    roi_80_twf_cls_type = os.path.join(twf_da_path, cls_type)
                    if os.path.exists(mask_path_cls_type) is False:
                        os.makedirs(mask_path_cls_type)
                    if os.path.exists(roi_path_cls_type) is False:
                        os.makedirs(roi_path_cls_type)
                    if os.path.exists(roi_corrected_cls_type) is False:
                        os.makedirs(roi_corrected_cls_type)
                    if os.path.exists(roi_80_afc_cls_type) is False:
                        os.makedirs(roi_80_afc_cls_type)
                    if os.path.exists(roi_80_H_cls_type) is False:
                        os.makedirs(roi_80_H_cls_type)
                    if os.path.exists(roi_80_PS_cls_type) is False:
                        os.makedirs(roi_80_PS_cls_type)
                    if os.path.exists(roi_80_norm_cls_type) is False:
                        os.makedirs(roi_80_norm_cls_type)
                    if os.path.exists(roi_80_twf_cls_type) is False:
                        os.makedirs(roi_80_twf_cls_type)

                X = []
                Y = []
                X_BV1 = []
                Y_BV1 = []
                i_reg = 0
                i_BV1 = []
                i_BV2 = []
                for Reg in Region:
                    print("Process Region No:", i_reg)
                    if (Reg.getAttribute("Text") == "BV1"):
                        i_BV1.append(i_reg)
                    if (Reg.getAttribute("Text") == "BV2"):
                        i_BV2.append(i_reg)
                    X.append([])
                    Y.append([])
                    X_BV1.append([])
                    Y_BV1.append([])
                    Vertex = Reg.getElementsByTagName("Vertex")
                    for Vert in Vertex:
                        X[i_reg].append(int(round(float(Vert.getAttribute("X")))))
                        Y[i_reg].append(int(round(float(Vert.getAttribute("Y")))))

                    i_reg += 1

                i1 = 0
                points = []
                for j in range(0, h, 2000):
                    for i in range(0, w, 2000):
                        print(os.path.join(self.tiles_dir, slide, 'Da' + str(i1) + ".jpg"))
                        img = cv2.imread(os.path.join(self.tiles_dir, slide, 'Da' + str(i1) + ".jpg"))
                        [hh, ww, cc] = img.shape
                        blank_image_all = np.zeros((hh, ww), np.uint8)
                        cnt_t = 0
                        for k in range(len(X)):
                            blank_image = np.zeros((hh, ww), np.uint8)
                            # print("######")
                            if i < max(X[k]) and i + 2000 > min(X[k]) and j < max(Y[k]) and j + 2000 > min(Y[k]):

                                points = []
                                for i3 in range(len(X[k])):
                                    points.append([int((X[k][i3] - i) / ds), int((Y[k][i3] - j) / ds)])
                                pts = np.array(points, np.int32)
                                pts = pts.reshape((-1, 1, 2))
                                cnt_t+=1
                                cv2.drawContours(blank_image, [pts], 0, (255), -1)
                                #cv2.drawContours(blank_image_all, [pts], 0, (255), -1)
                                cv2.imwrite(os.path.join(mask_path, class_types_reg[k], 'Da' + str(i1) + ".jpg"), blank_image)
                                bm_img = cv2.cvtColor(blank_image, cv2.COLOR_GRAY2BGR)
                                result_img = cv2.bitwise_and(img, bm_img)
                                invert_mask = cv2.bitwise_not(bm_img)

                                # Keep tiles with white background if they are not part of ROI ##### ROI_DA
                                da_or_img = cv2.bitwise_or(invert_mask, img)
                                cv2.imwrite(os.path.join(da_configpath, class_types_reg[k], 'Da' + str(i1) + ".jpg"), da_or_img)

                                ######################################### artifact filter #########################################
                                ################### Taking the da_or_img after multiply per ROI ###################################
                                mask_artifact_corrected = Artifact_SS1(da_or_img)[1]
                                total_pixels = mask_artifact_corrected.shape[0] * mask_artifact_corrected.shape[1]

                                # Count the number of white pixels
                                white_pixels = np.count_nonzero(mask_artifact_corrected)

                                #Calculate the percentage of white pixels
                                white_percentage = (white_pixels / total_pixels) * 100
                                #print(int(white_percentage))

                                if white_percentage < self.thresh_d:#80
                                    continue

                                cv2.imwrite(os.path.join(roi_corrected, class_types_reg[k], 'Da' + str(i1) + ".jpg"), mask_artifact_corrected)
                                cv2.imwrite(os.path.join(roi_corrected_da_path, class_types_reg[k], 'Da' + str(i1) + ".jpg"), da_or_img)
                                Inorm_img, PS_img, H_img = normalizeStaining(img, saveFile=None, Io=240, alpha=1, beta=0.15)

                                orig_PS = PS_img
                                bin_mask = cv2.cvtColor(orig_PS, cv2.COLOR_BGR2GRAY)

                                # Set the threshold value
                                threshold_value = 100  ####

                                # Create a mask of pixels less than the threshold value
                                mask = bin_mask < threshold_value
                                mask = mask * 255
                                mask = np.uint8(mask)
                                ### Add/Modify to 3channel if the mask is two channel
                                mask = cv2.cvtColor(mask, cv2.COLOR_GRAY2RGB)

                                da_and_img = cv2.bitwise_and(orig_PS, mask)
                                invert_mask = cv2.bitwise_not(mask)
                                roi_with_bg_img = cv2.bitwise_or(da_and_img, invert_mask)

                                cv2.imwrite(os.path.join(H_da_path, class_types_reg[k], 'Da' + str(i1) + ".jpg"), PS_img)
                                cv2.imwrite(os.path.join(PS_da_path, class_types_reg[k], 'Da' + str(i1) + ".jpg"), H_img)
                                cv2.imwrite(os.path.join(norm_da_path, class_types_reg[k], 'Da' + str(i1) + ".jpg"), Inorm_img)
                                cv2.imwrite(os.path.join(twf_da_path, class_types_reg[k], 'Da' + str(i1) + ".png"), roi_with_bg_img)
                        i1 += 1
            else: # process only if the output directories and results have not been processed and ready
                continue
def map_of_slides_and_annotations(input_slide_dir, wsi_tiles_dir, output_dir, file_type, thresh_d):

    params_keys = {'input_slide_dir': input_slide_dir,
                   'tiles_dir': wsi_tiles_dir,
                   'output_dir': output_dir,
                   'ext': file_type,
                   'thresh_d': thresh_d}
    obj = GenerateWSIannotation_on_cws(**params_keys)

    obj.generate_patch_of_annotated_tiles()

# slide_directory = r'D:\Projects\cbias-nap-AMY\slides'
# tiles_directory = r'D:\Projects\cbias-nap-AMY\cws'
# output_directory = r'D:\Projects\cbias-nap-AMY\refined_workflow'
# ext = '.ndpi'
#
# map_of_slides_and_annotations(slide_directory, tiles_directory, output_directory, ext)

# if __name__ == '__main__':
#     params = {'input_slide_dir': r'D:\AMY\230803_annotated_PSR\batch 1',  # input slide dir
#               'tiles_dir': r'D:\AMY\230803_PSR_ROI_tiles\cws',
#               'output_dir': r'D:\AMY\230803_PSR_ROI_tiles\refined_workflow',  # output dir 1.img_mask,2.Mat_files
#               'ext': '.ndpi',  # WSI format
#               }
#
#     obj = GenerateWSIannotation_on_cws(**params)
#     obj.generate_patch_of_annotated_tiles()
